from pathlib import Path, PosixPath
from typing import Dict, List, Literal, Optional, TypedDict, TypeVar, Union

PathOrStr = TypeVar("PathOrStr", Path, PosixPath, str)

NbNode = Dict[str, Union[str, int, "NbNode"]]
Metadata = Dict[str, Union[str, int, "Metadata"]]
MultilineText = Union[str, List[str]]


class NbMetadata(TypedDict):
    language_info: Metadata
    kernelspec: Metadata
    authors: Metadata


class Output(TypedDict):
    output_type: Literal[
        "execute_result",
        "display_data",
        "stream",
        "error",
    ]
    execution_count: Optional[int]
    metadata: Metadata


class ExecuteResult(Output):  # output_type = "execute_result"
    data: Dict[str, MultilineText]


class DisplayData(Output):  # output_type = "display_data"
    data: Dict[str, MultilineText]  # fix it - mimebundle


class Stream(Output):  # output_type = "stream"
    name: Literal["stdout", "stderr"]  # "The name of the stream (stdout, stderr)."
    text: MultilineText


class Error(Output):  # output_type = "error"
    ename: str  # "The name of the error."
    evalue: str  # "The value, or message, of the error."
    traceback: List[str]


class Cell(TypedDict):
    """Notebook cell base."""

    id: int  # from nbformat 4.5
    cell_type: Literal["code", "markdown", "raw"]
    metadata: Metadata
    source: MultilineText
    attachments: Optional[Dict[str, MultilineText]]


class CodeCell(Cell):  # cell_type = "code"
    """Code cell."""

    outputs: List[Output]
    execution_count: Optional[int]


class Nb(TypedDict):
    """Notebook."""

    nbformat: int
    nbformat_minor: int
    cells: List[Cell]
    metadata: Metadata



from pathlib import Path

from nbmetaclean.helpers import get_nb_names, is_notebook


def test_is_notebook():
    """test is_notebook"""
    assert is_notebook(Path("tests/test_nbs/test_nb_1.ipynb"))
    assert not is_notebook(Path("tests/test_nbs/test_nb_1.py"))
    assert not is_notebook(Path("tests/test_nbs/.test_nb_2_meta.ipynb"))
    assert is_notebook(Path("tests/test_nbs/.test_nb_2_meta.ipynb"), hidden=True)


def test_get_nb_names():
    """test get_nb_names"""
    path = Path("tests/test_nbs")
    # filename as argument
    file = path / "test_nb_1.ipynb"
    names = get_nb_names(file)
    assert len(names) == 1
    names.sort(key=lambda x: x.name)
    assert names[0] == file
    # filename but not nb
    names = get_nb_names("tests/test_clean.py")
    assert len(names) == 0

    # path as argument
    names = get_nb_names(path)
    assert len(names) == 2
    names.sort(key=lambda x: x.name)
    assert names[0] == file
    # path as argument. add hidden files
    names = get_nb_names(path, hidden=True)
    assert len(names) == 3
    try:
        get_nb_names("wrong_name")
        assert False
    except FileNotFoundError as ex:
        assert True
        assert str(ex) == "wrong_name not exists!"


def test_get_nb_names_recursive_hidden(tmp_path: Path):
    """test get_nb_names recursive hidden"""
    suffix = ".ipynb"
    # add one nb
    with open((tmp_path / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1

    # add hidden nb
    with open((tmp_path / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 2
    # add simple file
    with open((tmp_path / "simple"), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 1

    # add dir with one nb, hidden nb
    new_dir = tmp_path / "new_dir"
    new_dir.mkdir()
    with open((new_dir / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    with open((new_dir / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 2
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 4

    files = get_nb_names(tmp_path, recursive=False)
    assert len(files) == 1

    # add hidden dir
    hid_dir = tmp_path / ".hid_dir"
    hid_dir.mkdir()
    with open((hid_dir / "tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    with open((hid_dir / ".tst").with_suffix(suffix), "w", encoding="utf-8") as _:
        pass
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 6
    files = get_nb_names(tmp_path)
    assert len(files) == 2

    # add checkpoint dir and file
    # files at this dir will be skipped
    checkpoint_dir = tmp_path / ".ipynb_checkpoints"
    checkpoint_dir.mkdir()
    with open(
        (checkpoint_dir / "nb-checkpoint").with_suffix(suffix), "w", encoding="utf-8"
    ) as _:
        pass
    with open(
        (checkpoint_dir / "some_nb").with_suffix(suffix), "w", encoding="utf-8"
    ) as _:
        pass
    files = get_nb_names(tmp_path)
    assert len(files) == 2
    files = get_nb_names(tmp_path, hidden=True)
    assert len(files) == 6



import argparse
from pathlib import Path
from typing import Union

from .clean import CleanConfig, TupleStr, clean_nb_file
from .helpers import get_nb_names

parser = argparse.ArgumentParser(
    prog="nbclean",
    description="Clean metadata and execution_count from Jupyter notebooks.",
)
parser.add_argument(
    "path",
    default=".",
    nargs="*",
    help="Path for nb or folder with notebooks.",
)
parser.add_argument(
    "-s",
    "--silent",
    action="store_true",
    help="Silent mode.",
)
parser.add_argument(
    "--not_ec",
    action="store_false",
    help="Do not clear execution_count.",
)
parser.add_argument(
    "--not-pt",
    action="store_true",
    help="Do not preserve timestamp.",
)
parser.add_argument(
    "--dont_clear_nb_metadata",
    action="store_true",
    help="Do not clear notebook metadata.",
)
parser.add_argument(
    "--clear_cell_metadata",
    action="store_true",
    help="Clear cell metadata.",
)
parser.add_argument(
    "--clear_outputs",
    action="store_true",
    help="Clear outputs.",
)
parser.add_argument(
    "--nb_metadata_preserve_mask",
    nargs="+",
    help="Preserve mask for notebook metadata.",
)
parser.add_argument(
    "--cell_metadata_preserve_mask",
    nargs="+",
    help="Preserve mask for cell metadata.",
)
parser.add_argument(
    "--dont_merge_masks",
    action="store_true",
    help="Do not merge masks.",
)
parser.add_argument(
    "--clean_hidden_nbs",
    action="store_true",
    help="Clean hidden notebooks.",
)
parser.add_argument(
    "-D",
    "--dry_run",
    action="store_true",
    help="perform a trial run, don't write results",
)
parser.add_argument(
    "-V",
    "--verbose",
    action="store_true",
    help="Verbose mode. Print extra information.",
)


def process_mask(mask: Union[list[str], None]) -> Union[tuple[TupleStr, ...], None]:
    if mask is None:
        return None
    return tuple(tuple(item.split(".")) for item in mask)


def print_result(
    cleaned: list[Path],
    errors: list[tuple[Path, Exception]],
    clean_config: CleanConfig,
    path: list[Path],
    num_nbs: int,
) -> None:
    if clean_config.verbose:
        print(
            f"Path: {', '.join(path)}, preserve timestamp: {clean_config.preserve_timestamp}"
        )
        print(f"checked: {num_nbs} notebooks")
    if cleaned:
        if len(cleaned) == 1:
            print(f"cleaned: {cleaned[0]}")
        else:
            print(f"cleaned: {len(cleaned)} notebooks")
            for nb in cleaned:
                print("- ", nb)
    if errors:
        print(f"with errors: {len(errors)}")
        for nb, exc in errors:
            print(f"{nb}: {exc}")


def app() -> None:
    """Clean metadata and execution_count from Jupyter notebook."""
    cfg = parser.parse_args()
    clean_config = CleanConfig(
        clear_nb_metadata=not cfg.dont_clear_nb_metadata,
        clear_cell_metadata=cfg.clear_cell_metadata,
        clear_execution_count=cfg.not_ec,
        clear_outputs=cfg.clear_outputs,
        preserve_timestamp=not cfg.not_pt,
        silent=cfg.silent,
        nb_metadata_preserve_mask=process_mask(cfg.nb_metadata_preserve_mask),
        cell_metadata_preserve_mask=process_mask(cfg.cell_metadata_preserve_mask),
        mask_merge=not cfg.dont_merge_masks,
        dry_run=cfg.dry_run,
        verbose=cfg.verbose if not cfg.silent else False,
    )
    path_list = cfg.path if isinstance(cfg.path, list) else [cfg.path]
    nb_files: list[Path] = []
    for path in path_list:
        path = Path(path)
        if path.exists():
            nb_files.extend(get_nb_names(path, hidden=cfg.clean_hidden_nbs))
        else:
            print(f"{path} not exists!")

    cleaned, errors = clean_nb_file(
        nb_files,
        clean_config,
    )

    if not cfg.silent:
        print_result(cleaned, errors, clean_config, path_list, len(nb_files))


if __name__ == "__main__":
    app()



from pathlib import Path

from nbmetaclean.helpers import read_nb, write_nb


def test_read_nb():
    """test read notebook"""
    file = Path("tests/test_nbs/test_nb_1.ipynb")
    nb = read_nb(file)
    assert isinstance(nb, dict)
    assert nb["metadata"]["language_info"] == {"name": "python"}
    assert nb["metadata"]["authors"][0]["name"] == "Andrei Yasyrev"
    assert nb["nbformat"] == 4
    assert nb["nbformat_minor"] == 2
    cells = nb["cells"]
    assert isinstance(cells, list)
    assert len(cells) == 2
    # markdown
    assert cells[0]["cell_type"] == "markdown"
    assert cells[0]["source"] == []
    assert cells[0]["metadata"] == {}
    # code
    assert cells[1]["cell_type"] == "code"
    assert cells[1]["source"] == []
    assert cells[1]["execution_count"] is None
    assert cells[1]["metadata"] == {}
    assert cells[1]["outputs"] == []


def test_write_nb(tmp_path: Path):
    """test write notebook"""
    file = Path("tests/test_nbs/test_nb_1.ipynb")
    nb = read_nb(file)
    write_nb(nb, tmp_path / file.name)
    with open(tmp_path / file.name, "r", encoding="utf-8") as fh:
        res_text = fh.read()
    with open(file, "r", encoding="utf-8") as fh:
        org_text = fh.read()
    assert res_text == org_text

    # write with name w/o suffix
    result = write_nb(nb, tmp_path / "test_nb_1")
    assert result == tmp_path / "test_nb_1.ipynb"

    # write with stat
    stat = file.stat()
    timestamp = (stat.st_atime, stat.st_mtime)
    result = write_nb(nb, tmp_path / "test_nb_1", timestamp=timestamp)
    res_stat = result.stat()
    assert timestamp == (res_stat.st_atime, res_stat.st_mtime)



from setuptools import setup


REQUIREMENTS_FILENAME = "requirements.txt"
REQUIREMENTS_TEST_FILENAME = "requirements_test.txt"
REQUIREMENTS_DEV_FILENAME = "requirements_dev.txt"


def load_requirements(filename: str) -> list[str]:
    """Load requirements from file"""
    try:
        with open(filename, encoding="utf-8") as fh:
            return fh.read().splitlines()
    except FileNotFoundError:
        return []


REQUIRED = load_requirements(REQUIREMENTS_FILENAME)
TEST_REQUIRED = load_requirements(REQUIREMENTS_TEST_FILENAME)
DEV_REQUIRED = load_requirements(REQUIREMENTS_DEV_FILENAME)


# What packages are optional?
EXTRAS = {
    "test": TEST_REQUIRED,
    "dev": DEV_REQUIRED + TEST_REQUIRED,
}


setup(
    install_requires=REQUIRED,
    extras_require=EXTRAS,
)
